package colorspace

import "gonum.org/v1/gonum/mat"

type ChromaticAdaptation mat.Dense

func (a ChromaticAdaptation) AdaptXYZ(from, to Illuminant) mat.Matrix {
	var fromM, toM *mat.VecDense

	{
		Xw, Yw, Zw := from.ToXYZ()
		fromM = mat.NewVecDense(3, []float64{Xw, Yw, Zw})
	}

	{
		Xw, Yw, Zw := to.ToXYZ()
		toM = mat.NewVecDense(3, []float64{Xw, Yw, Zw})
	}

	var crdFrom, crdTo mat.VecDense

	crdFrom.MulVec((*mat.Dense)(&a), fromM)
	crdTo.MulVec((*mat.Dense)(&a), toM)

	return mat.NewDiagDense(3, []float64{
		crdTo.AtVec(0) / crdFrom.AtVec(0),
		crdTo.AtVec(1) / crdFrom.AtVec(1),
		crdTo.AtVec(2) / crdFrom.AtVec(2),
	})
}

var (
	ChromaticAdaptationBradford = mat.NewDense(3, 3, []float64{
		0.8951, 0.2664, -0.1614,
		-0.7502, 1.7135, 0.0367,
		0.0389, -0.0685, 1.0296,
	})
	ChromaticAdaptationCMCCAT2000 = mat.NewDense(3, 3, []float64{
		0.7982, 0.3389, -0.1371,
		-0.5918, 1.5512, 0.0406,
		0.0008, 0.0239, 0.9753,
	})
	ChromaticAdaptationCIECAT02 = mat.NewDense(3, 3, []float64{
		0.7328, 0.4296, -0.1624,
		-0.7036, 1.6975, 0.0061,
		0.0030, 0.0136, 0.9834,
	})
	ChromaticAdaptationSharp = mat.NewDense(3, 3, []float64{
		1.2694, -0.0988, -0.1706,
		-0.8364, 1.8006, 0.0357,
		0.0297, -0.0315, 1.0018,
	})
)
