package colorspace

import "gonum.org/v1/gonum/mat"

type Chromaticity struct {
	Red   ColorCoordinate
	Green ColorCoordinate
	Blue  ColorCoordinate
	White Illuminant
}

func (c Chromaticity) ConversionXYZ() (to, from *mat.Dense) {
	var err error
	var RGB *mat.Dense
	var W *mat.VecDense

	{
		Xr, Yr, Zr := c.Red.ToXYZ()
		Xg, Yg, Zg := c.Green.ToXYZ()
		Xb, Yb, Zb := c.Blue.ToXYZ()

		RGB = mat.NewDense(3, 3, []float64{
			Xr, Xg, Xb,
			Yr, Yg, Yb,
			Zr, Zg, Zb,
		})
	}

	{
		Xw, Yw, Zw := c.White.ToXYZ()

		W = mat.NewVecDense(3, []float64{Xw, Yw, Zw})
	}

	var tmp1, rgb2xyz, xyz2rgb mat.Dense

	var S mat.VecDense

	if err = tmp1.Inverse(RGB); err != nil {
		panic(err)
	}

	S.MulVec(&tmp1, W)

	M := mat.NewDense(3, 3, []float64{
		S.AtVec(0) * RGB.At(0, 0), S.AtVec(1) * RGB.At(0, 1), S.AtVec(2) * RGB.At(0, 2),
		S.AtVec(0) * RGB.At(1, 0), S.AtVec(1) * RGB.At(1, 1), S.AtVec(2) * RGB.At(1, 2),
		S.AtVec(0) * RGB.At(2, 0), S.AtVec(1) * RGB.At(2, 1), S.AtVec(2) * RGB.At(2, 2),
	})

	rgb2xyz.CloneFrom(M)

	if err = xyz2rgb.Inverse(M); err != nil {
		panic(err)
	}

	return &rgb2xyz, &xyz2rgb
}

/*
func (c Chromaticity) XYZToRGB(connectionSpaceWhite Illuminant, adaptation ChromaticAdaptation) {
	var err error
	var RGB *mat.Dense
	var W1, W2 *mat.VecDense

	{
		Xr, Yr, Zr := c.Red.ToXYZ()
		Xg, Yg, Zg := c.Green.ToXYZ()
		Xb, Yb, Zb := c.Blue.ToXYZ()

		RGB = mat.NewDense(3, 3, []float64{
			Xr, Xg, Xb,
			Yr, Yg, Yb,
			Zr, Zg, Zb,
		})
	}

	{
		Xw, Yw, Zw := c.White.ToXYZ()

		W1 = mat.NewVecDense(3, []float64{Xw, Yw, Zw})
	}

	{
		Xw, Yw, Zw := connectionSpaceWhite.ToXYZ()

		W2 = mat.NewVecDense(3, []float64{Xw, Yw, Zw})
	}

	var tmp1, tmp2, M, Mc, M2, rgb2xyz, xyz2rgb, source_destination_whites, destination_source_whites, adapted_rgb2xyz_2, adapted_xyz2rgb_2 mat.Dense

	var S, crdS, crdD, RA mat.VecDense

	if err = tmp1.Inverse(RGB); err != nil {
		panic(err)
	}

	S.MulVec(&tmp1, W1)

	//TODO
	M.Mul(&S, RGB)

	rgb2xyz.CloneFrom(&M)

	if err = xyz2rgb.Inverse(&M); err != nil {
		panic(err)
	}


	// chromatic adaptation

	crdS.MulVec((*mat.Dense)(&adaptation), W1)
	crdD.MulVec((*mat.Dense)(&adaptation), W2)

	Mt := mat.NewDiagDense(3, []float64{
		crdD.AtVec(0) / crdS.AtVec(0),
		crdD.AtVec(1) / crdS.AtVec(1),
		crdD.AtVec(2) / crdS.AtVec(2),
	})

	if err = tmp1.Inverse((*mat.Dense)(&adaptation)); err != nil {
		panic(err)
	}

	tmp2.Mul(&tmp1, Mt)
	Mc.Mul(&tmp2, (*mat.Dense)(&adaptation))

	source_destination_whites.CloneFrom(&Mc)

	if err = destination_source_whites.Inverse(&Mc); err != nil {
		panic(err)
	}

	M2.Mul(&Mc, &M)

	adapted_rgb2xyz_2.CloneFrom(&M2)

	if err = adapted_xyz2rgb_2.Inverse(&M2); err != nil {
		panic(err)
	}

	RA.MulVec(&Mc, W1)

}
*/
