package colorspace

type RelativeSystem struct {
	Chromaticity Chromaticity

	fromLinearTransfer TransferFunction

	YCbCr YCbCrConverter
}

func (s RelativeSystem) FromLinear(c float64) float64 {
	return s.fromLinearTransfer(c)
}

func NewRelativeSystem(chromaticity Chromaticity, fromLinearTransfer TransferFunction, converter YCbCrConverter) RelativeSystem {
	return RelativeSystem{
		Chromaticity:       chromaticity,
		fromLinearTransfer: fromLinearTransfer,
		YCbCr:              converter,
	}
}

var (
	SystemSRGB = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.640, 0.330},
		Green: ColorCoordinate{0.300, 0.600},
		Blue:  ColorCoordinate{0.150, 0.060},
		White: IlluminantD65,
	}, CompandingSRGB, YCbCr_Rec709)

	SystemRec709 = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.640, 0.330},
		Green: ColorCoordinate{0.300, 0.600},
		Blue:  ColorCoordinate{0.150, 0.060},
		White: IlluminantD65,
	}, CompandingRec709, YCbCr_Rec709)

	SystemRec709_Pure = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.640, 0.330},
		Green: ColorCoordinate{0.300, 0.600},
		Blue:  ColorCoordinate{0.150, 0.060},
		White: IlluminantD65,
	}, PureRec709, YCbCr_Rec709)

	SystemRec709_Pure22 = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.640, 0.330},
		Green: ColorCoordinate{0.300, 0.600},
		Blue:  ColorCoordinate{0.150, 0.060},
		White: IlluminantD65,
	}, PureRec709_22, YCbCr_Rec709)

	SystemRec709_Pure24 = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.640, 0.330},
		Green: ColorCoordinate{0.300, 0.600},
		Blue:  ColorCoordinate{0.150, 0.060},
		White: IlluminantD65,
	}, PureRec709_24, YCbCr_Rec709)

	SystemRec2020 = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.708, 0.292},
		Green: ColorCoordinate{0.170, 0.797},
		Blue:  ColorCoordinate{0.131, 0.046},
		White: IlluminantD65,
	}, CompandingRec2020, YCbCr_Rec2020)

	SystemRec2020_Pure = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.708, 0.292},
		Green: ColorCoordinate{0.170, 0.797},
		Blue:  ColorCoordinate{0.131, 0.046},
		White: IlluminantD65,
	}, PureRec2020, YCbCr_Rec2020)

	SystemRec2020_Pure24 = NewRelativeSystem(Chromaticity{
		Red:   ColorCoordinate{0.708, 0.292},
		Green: ColorCoordinate{0.170, 0.797},
		Blue:  ColorCoordinate{0.131, 0.046},
		White: IlluminantD65,
	}, PureRec2020, YCbCr_Rec2020)
)
