#include "conv.h"

#include <math.h>
#include <stdlib.h>

const double uint16Max = (double)((1<<16)-1);
const double int16MaxPlusOne = (double)(1<<15);

const double* restrict xyz2rgb_g;
const double* restrict rgb2yuv_g;
double rgbGamma_g;


#if USE_SIMD
#include "conv_gen.h"
#endif

#if USE_SIMD && __AVX2__
#include "conv_avx.h"

#if USE_512_WIDE_PIPELINE && __AVX512F__ && __AVX512VL__ && __AVX512BW__
//512-bit wide pipeline
typed_converter_platform(x86_64, SIMD_PUMPS, float, 5, 1, load_packed_512f, store_packed_512f);
typed_converter_platform(x86_64, SIMD_PUMPS, double, 2, 1, load_packed_512, store_packed_512);

const char* DecoderInformation = "SIMD AVX-512 512-bit pipeline (2d 5f " str(SIMD_PUMPS) "pump)";
#else

#if !defined(__FMA__)
#define __FMA__ 0
const char* DecoderInformation = "SIMD AVX2 256-bit pipeline (1d 2f " str(SIMD_PUMPS) "pump)";
#else
const char* DecoderInformation = "SIMD AVX2 + FMA 256-bit pipeline (1d 2f " str(SIMD_PUMPS) "pump)";
#endif

//256-bit wide pipeline
typed_converter_platform(x86_64, SIMD_PUMPS, float, 2, __FMA__, load_packed_256f, store_packed_256f);
typed_converter_platform(x86_64, SIMD_PUMPS, double, 1, __FMA__, load_packed_256, store_packed_256);

#endif

#else
//No ASM defined
#define fType float
#include "conv_generic.h"
#undef fType
#define fType double
#include "conv_generic.h"

#undef USE_SIMD
#define USE_SIMD 0

const char* DecoderInformation = "Generic scalar pipeline (1d 1f 1pump)";


#endif

const char* decoder_information() {
    return DecoderInformation;
}


void init(const double* restrict xyz2rgb, const double* restrict rgb2yuv, double xyzGamma, double rgbGamma) {
    xyz2rgb_g = xyz2rgb;
    rgb2yuv_g = rgb2yuv;
    rgbGamma_g = 1./rgbGamma;

    for (int i = 0; i < (1 << XYZ_LOOKUP_TABLE_SIZE); i++) {
        xyz12_to_linear_double[i] = pow((double)(i)/((1 << XYZ_LOOKUP_TABLE_SIZE)-1), xyzGamma);
        xyz12_to_linear_float[i] = xyz12_to_linear_double[i];
    }

    #if USE_SIMD
        _load_matrix_double(xyz2rgb_mat_double, xyz2rgb);
        _load_matrix_float(xyz2rgb_mat_float, xyz2rgb);
        _load_matrix_double(rgb2yuv_mat_double, rgb2yuv);
        _load_matrix_float(rgb2yuv_mat_float, rgb2yuv);
    #endif
}

void convert_frame_dci_xyz12_to_yuv16_float(const uint32_t* restrict in, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr, int width, int height) {
    for (int i = 0; i < height; i++){
        convert_line_dci_xyz12_to_yuv16_float(in, luma, cb, cr, width, height);

        in += width*3;

        luma += width;
        cb += width;
        cr += width;
    }
}

void convert_frame_dci_xyz12_to_yuv16_double(const uint32_t* restrict in, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr, int width, int height) {
    for (int i = 0; i < height; i++){
        convert_line_dci_xyz12_to_yuv16_double(in, luma, cb, cr, width, height);

        in += width*3;

        luma += width;
        cb += width;
        cr += width;
    }
}