package conv

import (
	"gonum.org/v1/gonum/mat"
	"runtime"
	"unsafe"
)

/*
#cgo CFLAGS: -march=native -mtune=native -Ofast -std=c99
#cgo LDFLAGS: -lm
#include "conv.h"
*/
import "C"

var matPinner runtime.Pinner

func DecoderInformation() string {
	return C.GoString(C.decoder_information())
}

func InitData(xyz2rgb, rgb2yuv *mat.Dense, xyzGamma, rgbGamma float64) {
	a := unsafe.Pointer(unsafe.SliceData(xyz2rgb.RawMatrix().Data))
	b := unsafe.Pointer(unsafe.SliceData(rgb2yuv.RawMatrix().Data))
	matPinner.Pin(a)
	matPinner.Pin(b)

	C.init((*C.double)(a), (*C.double)(b), C.double(xyzGamma), C.double(rgbGamma))
}

func ConvertFrameDCIXYZToYUV16Double(in []uint32, luma, cb, cr []uint16, width, height int) {

	var pinner runtime.Pinner

	inp := unsafe.Pointer(unsafe.SliceData(in))
	lumap := unsafe.Pointer(unsafe.SliceData(luma))
	cbp := unsafe.Pointer(unsafe.SliceData(cb))
	crp := unsafe.Pointer(unsafe.SliceData(cr))

	pinner.Pin(inp)
	pinner.Pin(lumap)
	pinner.Pin(cbp)
	pinner.Pin(crp)
	defer pinner.Unpin()

	C.convert_frame_dci_xyz12_to_yuv16_double(
		(*C.uint)(inp),
		(*C.ushort)(lumap),
		(*C.ushort)(cbp),
		(*C.ushort)(crp),
		C.int(width),
		C.int(height),
	)
}

func ConvertLineDCIXYZToYUV16Double(in []uint32, luma, cb, cr []uint16, width, height int) {
	var pinner runtime.Pinner

	inp := unsafe.Pointer(unsafe.SliceData(in))
	lumap := unsafe.Pointer(unsafe.SliceData(luma))
	cbp := unsafe.Pointer(unsafe.SliceData(cb))
	crp := unsafe.Pointer(unsafe.SliceData(cr))

	pinner.Pin(inp)
	pinner.Pin(lumap)
	pinner.Pin(cbp)
	pinner.Pin(crp)
	defer pinner.Unpin()

	C.convert_line_dci_xyz12_to_yuv16_double(
		(*C.uint)(inp),
		(*C.ushort)(lumap),
		(*C.ushort)(cbp),
		(*C.ushort)(crp),
		C.int(width),
		C.int(height),
	)
}

func ConvertFrameDCIXYZToYUV16Float(in []uint32, luma, cb, cr []uint16, width, height int) {

	var pinner runtime.Pinner

	inp := unsafe.Pointer(unsafe.SliceData(in))
	lumap := unsafe.Pointer(unsafe.SliceData(luma))
	cbp := unsafe.Pointer(unsafe.SliceData(cb))
	crp := unsafe.Pointer(unsafe.SliceData(cr))

	pinner.Pin(inp)
	pinner.Pin(lumap)
	pinner.Pin(cbp)
	pinner.Pin(crp)
	defer pinner.Unpin()

	C.convert_frame_dci_xyz12_to_yuv16_float(
		(*C.uint)(inp),
		(*C.ushort)(lumap),
		(*C.ushort)(cbp),
		(*C.ushort)(crp),
		C.int(width),
		C.int(height),
	)
}

func ConvertLineDCIXYZToYUV16Float(in []uint32, luma, cb, cr []uint16, width, height int) {
	var pinner runtime.Pinner

	inp := unsafe.Pointer(unsafe.SliceData(in))
	lumap := unsafe.Pointer(unsafe.SliceData(luma))
	cbp := unsafe.Pointer(unsafe.SliceData(cb))
	crp := unsafe.Pointer(unsafe.SliceData(cr))

	pinner.Pin(inp)
	pinner.Pin(lumap)
	pinner.Pin(cbp)
	pinner.Pin(crp)
	defer pinner.Unpin()

	C.convert_line_dci_xyz12_to_yuv16_float(
		(*C.uint)(inp),
		(*C.ushort)(lumap),
		(*C.ushort)(cbp),
		(*C.ushort)(crp),
		C.int(width),
		C.int(height),
	)
}
