#include <stdint.h>

void init(const double* restrict xyz2rgb, const double* restrict rgb2yuv, double xyzGamma, double rgbGamma);


// Use available SIMD. Disable to enforce generic pipeline.
#if !defined(USE_SIMD)
#define USE_SIMD 1
#endif

// Opportunistically use AVX512 features even in 256-bit mode
#if !defined(USE_OPPORTUNISTIC_AVX512)
#define USE_OPPORTUNISTIC_AVX512 1
#endif

//AVX2
// double layout aaa0
// float layout aaabbb00

// Enable usage of 512-bit wide pipeline, pumping two pixels every iteration, if supported.
// Uses AVX-512 features. Requires AVX-512 F, VL, BW. Layout aaabbb00
// double layout aaabbb00
// float layout aaabbbcccdddeee0
#if !defined(USE_512_WIDE_PIPELINE)
#define USE_512_WIDE_PIPELINE 1
#endif

// Sets the number of pumps per iteration on pipeline. Supported 1, 2, 4, 8
// Set this if your architecture has large amount of executors than normal. Recommended to stay at 2 or 4.
#if !defined(SIMD_PUMPS)
#define SIMD_PUMPS 1
#endif

// ExpandLoad or CompressStore is slower than currently doing set(vals...). todo: inspect with newer CPUs than ZEN4
#if !defined(USE_AVX512_EXPANDLOAD)
#define USE_AVX512_EXPANDLOAD 0
#endif

// Size of the lookup table. Only valid value is 12.
#if !defined(XYZ_LOOKUP_TABLE_SIZE)
#define XYZ_LOOKUP_TABLE_SIZE 12
#endif

// This function can run out of bounds slightly, about 24 bytes per pump per line (XYZ) or 8 bytes per pump per line (YUV), caller should allocates extra input/output buffer for this reason.
void convert_line_dci_xyz12_to_yuv16_float(const uint32_t* restrict in, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr, int width, int height);
void convert_frame_dci_xyz12_to_yuv16_float(const uint32_t* restrict in, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr, int width, int height);
void convert_line_dci_xyz12_to_yuv16_double(const uint32_t* restrict in, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr, int width, int height);
void convert_frame_dci_xyz12_to_yuv16_double(const uint32_t* restrict in, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr, int width, int height);

const char* decoder_information();