#include <immintrin.h>
#include <stdint.h>

// x86_64 AVX2 and AVX512 definitions



#define _perm_component_i(i, c, v) ((i*v+c)*2)
#define _perm_component2(c) _perm_component_i(1, c, 4), _perm_component_i(0, c, 4)
#define _perm_component5(c) _perm_component_i(4, c, 3), _perm_component_i(3, c, 3), _perm_component_i(2, c, 3), _perm_component_i(1, c, 3), _perm_component_i(0, c, 3)

#if USE_512_WIDE_PIPELINE && __AVX512F__ && __AVX512VL__ && __AVX512BW__ && __AVX512DQ__

__m256i load_packed_512(int j, const uint32_t* restrict in) {
#if USE_AVX512_EXPANDLOAD
    //todo: this path is slower than _mm256_set_epi32, both mask and maskz ZEN4
    return _mm256_maskz_expandloadu_epi32(_cvtu32_mask8(0b01110111), in+j*3);
#else
    return _mm256_permutexvar_epi32(_mm256_set_epi32(7, 5, 4, 3, 7, 2, 1, 0), _mm256_maskz_loadu_epi32(_cvtu32_mask8(0b00111111), in+j*3));
#endif
}

void store_packed_512(__m256i packed, int j, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr) {
    #define _perm_idx512 _mm256_set_epi16(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, _perm_component2(2), _perm_component2(1), _perm_component2(0))

    __m128i store = _mm256_castsi256_si128(_mm256_permutexvar_epi16(_perm_idx512, packed));
    _mm_mask_storeu_epi16(&luma[j], _cvtu32_mask8(0b00000011), store);
    _mm_mask_storeu_epi16(&cb[j-2], _cvtu32_mask8(0b00001100), store);
    _mm_mask_storeu_epi16(&cr[j-4], _cvtu32_mask8(0b00110000), store);
}


__m512i load_packed_512f(int j, const uint32_t* restrict in) {
    return _mm512_maskz_loadu_epi32(_cvtu32_mask16(0b0111111111111111), in+j*3);
}

void store_packed_512f(__m512i packed, int j, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr) {
    #define _perm_idx512f _mm512_set_epi16(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, _perm_component5(2), _perm_component5(1), _perm_component5(0))

    __m256i store = _mm512_castsi512_si256(_mm512_permutexvar_epi16(_perm_idx512f, packed));
    _mm256_mask_storeu_epi16(&luma[j ], _cvtu32_mask16(0b0000000000011111), store);
    _mm256_mask_storeu_epi16(&cb[j -5], _cvtu32_mask16(0b0000001111100000), store);
    _mm256_mask_storeu_epi16(&cr[j-10], _cvtu32_mask16(0b0111110000000000), store);
}

#define SSE 1
#define AVX 1
#define AVX512 1
#include "simd_utils/simd_utils.h"

static inline __m512d pow512_pd1(__m512d x, double y1) {
    const __m512d y = _mm512_set1_pd(y1);
    return exp512_pd(_mm512_mul_pd(y, log512_pd(x)));
}

static inline __m512 pow512_ps1(__m512 x, double y1) {
    const __m512d y = _mm512_set1_pd(y1);
    //low precision cause issues, use doubles
    __m256 a = _mm512_cvtpd_ps(exp512_pd(_mm512_mul_pd(y, log512_pd(_mm512_cvtps_pd(_mm512_castps512_ps256(x))))));
    __m256 b = _mm512_cvtpd_ps(exp512_pd(_mm512_mul_pd(y, log512_pd(_mm512_cvtps_pd(_mm512_extractf32x8_ps(x, 1))))));

    return _mm512_insertf32x8(_mm512_castps256_ps512(a), b, 1);
}

#else

__m128i load_packed_256(int j, const uint32_t* restrict in) {
#if USE_OPPORTUNISTIC_AVX512 && __AVX512F__ && __AVX512VL__ && __AVX512DQ__
    return _mm_maskz_loadu_epi32(_cvtu32_mask8(0b00000111), in+j*3);
#else
    return _mm_loadu_epi32(in+j*3);
    //return _mm_set_epi32(0, inz[j], iny[j], inx[j]);
#endif
}

void store_packed_256(__m128i packed, int j, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr) {
    luma[j] = ((int32_t*)(&packed))[0];
    cb[j] = ((int32_t*)(&packed))[1];
    cr[j] = ((int32_t*)(&packed))[2];
}

/*#if USE_DOT_PRODUCT
//position entries properly in vectors ready for dot product
m[0] = _mm256_set_ps(0, in[2], in[1], in[0], 0, in[2], in[1], in[0]); //0rrr 0rrr
m[1] = _mm256_set_ps(0, in[5], in[4], in[3], 0, in[5], in[4], in[3]); //0ggg 0ggg
m[2] = _mm256_set_ps(0, in[8], in[7], in[6], 0, in[8], in[7], in[6]); //0bbb 0bbb

    __m256 a = _mm256_dp_ps(v, m0, 0b01110001);
    __m256 b = _mm256_dp_ps(v, m1, 0b01110010);
    __m256 c = _mm256_dp_ps(v, m2, 0b01110100);
    return _mm256_blend_ps(_mm256_blend_ps(a, b, 0b00100010), c, 0b01000100);
*/

__m256i load_packed_256f(int j, const uint32_t* restrict in) {
#if USE_OPPORTUNISTIC_AVX512 && USE_AVX512_EXPANDLOAD && __AVX512VBMI2__ && __AVX512VL__ && __AVX512F__
    //todo: slow ZEN4
    return _mm256_maskz_expandloadu_epi32(_cvtu32_mask8(0b01110111), in+j*3);
#else
    #if USE_OPPORTUNISTIC_AVX512 && __AVX512F__ && __AVX512BW__ && __AVX512VL__ && __AVX512DQ__
    return _mm256_permutexvar_epi32(_mm256_set_epi32(7, 5, 4, 3, 7, 2, 1, 0), _mm256_maskz_loadu_epi32(_cvtu32_mask8(0b00111111), in+j*3));
    #else
    return _mm256_set_epi32(0, in[j+5], in[j+4], in[j+3], 0, in[j+2], in[j+1], in[j]);
    #endif
#endif
}

void store_packed_256f(__m256i packed, int j, uint16_t* restrict luma, uint16_t* restrict cb, uint16_t* restrict cr) {
#if USE_OPPORTUNISTIC_AVX512 && __AVX512BW__ && __AVX512VL__ && __AVX512DQ__

#define _perm_idx256 _mm256_set_epi16(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, _perm_component2(2), _perm_component2(1), _perm_component2(0))

__m128i store = _mm256_castsi256_si128(_mm256_permutexvar_epi16(_perm_idx256, packed));
_mm_mask_storeu_epi16(&luma[j], _cvtu32_mask8(0b00000011), store);
_mm_mask_storeu_epi16(&cb[j-2], _cvtu32_mask8(0b00001100), store);
_mm_mask_storeu_epi16(&cr[j-4], _cvtu32_mask8(0b00110000), store);

#else
    luma[j/3] = ((int32_t*)(&packed))[0];
    cb[j/3] = ((int32_t*)(&packed))[1];
    cr[j/3] = ((int32_t*)(&packed))[2];

    luma[j/3+1] = ((int32_t*)(&packed))[4];
    cb[j/3+1] = ((int32_t*)(&packed))[5];
    cr[j/3+1] = ((int32_t*)(&packed))[6];

#endif
}

#define SSE 1
#define AVX 1
#include "simd_utils/simd_utils.h"

static inline __m256d pow256_pd1(__m256d x, double y1) {
    const __m256d y = _mm256_set1_pd(y1);
    return exp256_pd(_mm256_mul_pd(y, log256_pd(x)));
}

static inline __m256 pow256_ps1(__m256 x, double y1) {
    const __m256d y = _mm256_set1_pd(y1);
    //low precision cause issues, use doubles
    __m128 a = _mm256_cvtpd_ps(exp256_pd(_mm256_mul_pd(y, log256_pd(_mm256_cvtps_pd(_mm256_castps256_ps128(x))))));
    __m128 b = _mm256_cvtpd_ps(exp256_pd(_mm256_mul_pd(y, log256_pd(_mm256_cvtps_pd(_mm256_extractf128_ps(x, 1))))));

    return _mm256_insertf128_ps(_mm256_castps128_ps256(a), b, 1);
}

#endif




#define typed_vector_size_x86_64_float_5 512
#define typed_vector_size_x86_64_double_2 512
#define typed_vector_size_x86_64_float_2 256
#define typed_vector_size_x86_64_double_1 256

#define typed_permute_lanes_x86_64_float_5 _permutexvar_
#define typed_permute_lanes_x86_64_double_2 _permutex_
#define typed_permute_lanes_x86_64_float_2 _permute_
#define typed_permute_lanes_x86_64_double_1 _permute4x64_


inline __attribute__((always_inline)) __m512 _int_i32gather_ps(void const* base_addr, __m512i vindex, int scale) {
    switch(scale){
        case 1:
        return _mm512_i32gather_ps(vindex, base_addr, 1);
        case 2:
        return _mm512_i32gather_ps(vindex, base_addr, 2);
        case 4:
        return _mm512_i32gather_ps(vindex, base_addr, 4);
        case 8:
        return _mm512_i32gather_ps(vindex, base_addr, 8);
        default:
        __builtin_unreachable();
    }
}

inline __attribute__((always_inline)) __m512d _int_i32gather_pd(void const* base_addr, __m256i vindex, int scale) {
    switch(scale){
        case 1:
        return _mm512_i32gather_pd(vindex, base_addr, 1);
        case 2:
        return _mm512_i32gather_pd(vindex, base_addr, 2);
        case 4:
        return _mm512_i32gather_pd(vindex, base_addr, 4);
        case 8:
        return _mm512_i32gather_pd(vindex, base_addr, 8);
        default:
        __builtin_unreachable();
    }
}

#define typed_pow1_x86_64_float_5 pow512_ps1
#define typed_pow1_x86_64_double_2 pow512_pd1
#define typed_pow1_x86_64_float_2 pow256_ps1
#define typed_pow1_x86_64_double_1 pow256_pd1
#define typed_pow1_x86_64(floatType, elementCount) typed_pow1_x86_64_##floatType##_##elementCount

#define typed_i32gather_x86_64_float_5 _int_i32gather_ps
#define typed_i32gather_x86_64_double_2 _int_i32gather_pd
#define typed_i32gather_x86_64_float_2 _mm256_i32gather_ps
#define typed_i32gather_x86_64_double_1 _mm256_i32gather_pd
#define typed_i32gather_x86_64(floatType, elementCount) typed_i32gather_x86_64_##floatType##_##elementCount

#define typed_vector_func_prefix_x86_64 _mm
#define typed_vector_type_prefix_x86_64 __m
#define typed_vector_type_suffix_x86_64_double d
#define typed_vector_type_suffix_x86_64_int i

#define typed_vector_int_size_x86_64_float_5 512
#define typed_vector_int_func_x86_64_float_5 concat(typed_vector_func_prefix_x86_64, typed_vector_int_size_x86_64_float_5)
#define typed_vector_int_size_x86_64_double_2 256
#define typed_vector_int_func_x86_64_double_2 concat(typed_vector_func_prefix_x86_64, typed_vector_int_size_x86_64_double_2)
#define typed_vector_int_size_x86_64_float_2 256
#define typed_vector_int_func_x86_64_float_2 concat(typed_vector_func_prefix_x86_64, typed_vector_int_size_x86_64_float_2)
#define typed_vector_int_size_x86_64_double_1 128
#define typed_vector_int_func_x86_64_double_1 typed_vector_func_prefix_x86_64

#define typed_vector_func_type_x86_64_float ps
#define typed_vector_func_type_x86_64_double pd

#define typed_vector_type_x86_64_float(elementCount) concat(typed_vector_type_prefix_x86_64, typed_vector_size_x86_64_float_##elementCount)
#define typed_vector_type_x86_64_double(elementCount) concat3(typed_vector_type_prefix_x86_64, typed_vector_size_x86_64_double_##elementCount, typed_vector_type_suffix_x86_64_double)

#define typed_vector_type_x86_64(floatType, elementCount) typed_vector_type_x86_64_##floatType(elementCount)

#define typed_vector_int_type_x86_64(floatType, elementCount) concat3(typed_vector_type_prefix_x86_64, typed_vector_int_size_x86_64_##floatType##_##elementCount, typed_vector_type_suffix_x86_64_int)

#define typed_vector_func_x86_64_float(elementCount) concat(typed_vector_func_prefix_x86_64, typed_vector_size_x86_64_float_##elementCount)
#define typed_vector_func_x86_64_double(elementCount) concat(typed_vector_func_prefix_x86_64, typed_vector_size_x86_64_double_##elementCount)

#define typed_vector_func_x86_64(floatType, elementCount) typed_vector_func_x86_64_##floatType(elementCount)



#define typed_permute_lanes_x86_64_float_5 _permutexvar_
#define typed_permute_lanes_x86_64_double_2 _permutex_
#define typed_permute_lanes_x86_64_float_2 _permute_
#define typed_permute_lanes_x86_64_double_1 _permute4x64_

#define typed_fmadd_x86_64 _fmadd_
#define typed_add_x86_64 _add_
#define typed_mul_x86_64 _mul_
#define typed_set1_x86_64 _set1_
#define typed_set_x86_64 _set_
#define typed_min_x86_64 _min_
#define typed_max_x86_64 _max_
#define typed_seti_x86_64 _set_epi32
#define typed_addi_x86_64 _add_epi32
#define typed_cvt_x86_64_float _cvtps_epi32
#define typed_cvt_x86_64_double _cvtpd_epi32

#define typed_func_x86_64_fmadd(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_fmadd_x86_64, typed_vector_func_type_x86_64_##floatType)
#define typed_func_x86_64_add(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_add_x86_64, typed_vector_func_type_x86_64_##floatType)
#define typed_func_x86_64_mul(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_mul_x86_64, typed_vector_func_type_x86_64_##floatType)
#define typed_func_x86_64_set1(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_set1_x86_64, typed_vector_func_type_x86_64_##floatType)
#define typed_func_x86_64_set(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_set_x86_64, typed_vector_func_type_x86_64_##floatType)
#define typed_func_x86_64_min(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_min_x86_64, typed_vector_func_type_x86_64_##floatType)
#define typed_func_x86_64_max(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_max_x86_64, typed_vector_func_type_x86_64_##floatType)
#define typed_func_x86_64_ftoi(floatType, elementCount) concat(typed_vector_func_x86_64(floatType, elementCount), typed_cvt_x86_64_##floatType)
#define typed_func_x86_64_seti(floatType, elementCount) concat(typed_vector_int_func_x86_64_##floatType##_##elementCount, typed_seti_x86_64)
#define typed_func_x86_64_addi(floatType, elementCount) concat(typed_vector_int_func_x86_64_##floatType##_##elementCount, typed_addi_x86_64)

#define typed_func_x86_64_pow1(floatType, elementCount) typed_pow1_x86_64(floatType, elementCount)
#define typed_func_x86_64_i32gather(floatType, elementCount) typed_i32gather_x86_64(floatType, elementCount)
#define typed_func_x86_64_permute_lanes(floatType, elementCount) concat3(typed_vector_func_x86_64(floatType, elementCount), typed_permute_lanes_x86_64_##floatType##_##elementCount, typed_vector_func_type_x86_64_##floatType)

#define typed_permute_lanes_x86_64_float_5 _permutexvar_
#define typed_permute_lanes_x86_64_double_2 _permutex_
#define typed_permute_lanes_x86_64_float_2 _permute_
#define typed_permute_lanes_x86_64_double_1 _permute4x64_
